@@ -40,12 +40,14 @@ Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 #include "insn-attr.h"
 #include "c-tree.h"
 #include "optabs.h"
+#include "reload.h"
 #include "protector.h"
 
 
 void prepare_stack_protection  PARAMS ((int inlinable));
 int search_string_def PARAMS ((tree names));
-rtx assign_stack_local_for_pseudo_reg PARAMS ((enum machine_mode, HOST_WIDE_INT, int));
+rtx assign_stack_local_for_pseudo_reg PARAMS ((enum machine_mode,
+					       HOST_WIDE_INT, int));
 
 
 /* Warn when not issuing stack smashing protection for some reason */
@@ -76,6 +78,7 @@ static int current_function_defines_short_string;
 static int current_function_has_variable_string;
 static int current_function_defines_vsized_array;
 static int current_function_is_inlinable;
+static int is_array;
 
 static rtx guard_area, _guard;
 static rtx function_first_insn, prologue_insert_point;
@@ -89,7 +92,7 @@ static int saved_cse_not_expected = 0;
 static int search_string_from_argsandvars PARAMS ((int caller));
 static int search_string_from_local_vars PARAMS ((tree block));
 static int search_pointer_def PARAMS ((tree names));
-static int search_func_pointer PARAMS ((tree type, int mark));
+static int search_func_pointer PARAMS ((tree type));
 static void reset_used_flags_for_insns PARAMS ((rtx insn));
 static void reset_used_flags_for_decls PARAMS ((tree block));
 static void reset_used_flags_of_plus PARAMS ((rtx x));
@@ -97,15 +100,25 @@ static void rtl_prologue PARAMS ((rtx insn));
 static void rtl_epilogue PARAMS ((rtx fnlastinsn));
 static void arrange_var_order PARAMS ((tree blocks));
 static void copy_args_for_protection PARAMS ((void));
-static void sweep_string_variable PARAMS ((rtx sweep_var, HOST_WIDE_INT var_size));
-static void sweep_string_in_decls PARAMS ((tree block, HOST_WIDE_INT sweep_offset, HOST_WIDE_INT size));
-static void sweep_string_in_args PARAMS ((tree parms, HOST_WIDE_INT sweep_offset, HOST_WIDE_INT size));
-static void sweep_string_use_of_insns PARAMS ((rtx insn, HOST_WIDE_INT sweep_offset, HOST_WIDE_INT size));
-static void sweep_string_in_operand PARAMS ((rtx insn, rtx *loc, HOST_WIDE_INT sweep_offset, HOST_WIDE_INT size));
-static void move_arg_location PARAMS ((rtx insn, rtx orig, rtx new, HOST_WIDE_INT var_size));
-static void change_arg_use_of_insns PARAMS ((rtx insn, rtx orig, rtx *new, HOST_WIDE_INT size));
-static void change_arg_use_of_insns_2 PARAMS ((rtx insn, rtx orig, rtx *new, HOST_WIDE_INT size));
-static void change_arg_use_in_operand PARAMS ((rtx x, rtx orig, rtx *new, HOST_WIDE_INT size));
+static void sweep_string_variable
+	PARAMS ((rtx sweep_var, HOST_WIDE_INT var_size));
+static void sweep_string_in_decls
+	PARAMS ((tree block, HOST_WIDE_INT sweep_offset, HOST_WIDE_INT size));
+static void sweep_string_in_args
+	PARAMS ((tree parms, HOST_WIDE_INT sweep_offset, HOST_WIDE_INT size));
+static void sweep_string_use_of_insns
+	PARAMS ((rtx insn, HOST_WIDE_INT sweep_offset, HOST_WIDE_INT size));
+static void sweep_string_in_operand
+	PARAMS ((rtx insn, rtx *loc,
+		 HOST_WIDE_INT sweep_offset, HOST_WIDE_INT size));
+static void move_arg_location
+	PARAMS ((rtx insn, rtx orig, rtx new, HOST_WIDE_INT var_size));
+static void change_arg_use_of_insns
+	PARAMS ((rtx insn, rtx orig, rtx *new, HOST_WIDE_INT size));
+static void change_arg_use_of_insns_2
+	PARAMS ((rtx insn, rtx orig, rtx *new, HOST_WIDE_INT size));
+static void change_arg_use_in_operand
+	PARAMS ((rtx x, rtx orig, rtx *new, HOST_WIDE_INT size));
 static void validate_insns_of_varrefs PARAMS ((rtx insn));
 static void validate_operand_of_varrefs PARAMS ((rtx insn, rtx *loc));
 
@@ -120,6 +133,7 @@ static void validate_operand_of_varrefs PARAMS ((rtx insn, rtx *loc));
  (GET_CODE (DECL_INCOMING_RTL (PARM)) == MEM)
 #define VIRTUAL_STACK_VARS_P(X) \
  ((X) == virtual_stack_vars_rtx || (GET_CODE (X) == REG && (X)->used))
+#define TREE_VISITED(NODE) ((NODE)->common.unused_0)
 
 
 
@@ -133,14 +147,17 @@ prepare_stack_protection (inlinable)
   saved_cse_not_expected = 0;
 
   /*
-    skip the protection if the function has no block or it is an inline function
+    skip the protection if the function has no block
+    or it is an inline function
   */
   if (current_function_is_inlinable) validate_insns_of_varrefs (get_insns ());
   if (! blocks || current_function_is_inlinable) return;
 
-  current_function_defines_vulnerable_string = search_string_from_argsandvars (0);
+  current_function_defines_vulnerable_string
+    = search_string_from_argsandvars (0);
 
-  if (current_function_defines_vulnerable_string)
+  if (current_function_defines_vulnerable_string
+      || flag_stack_protection)
     {
       HOST_WIDE_INT offset;
       function_first_insn = get_insns ();
@@ -171,16 +188,17 @@ prepare_stack_protection (inlinable)
       sweep_frame_offset = frame_offset;
 #endif
 
-      /* For making room for guard value, scan all insns and fix the offset address
-	 of the variable that is based on frame pointer.
-	 Scan all declarations of variables and fix the offset address of the variable that
-	 is based on the frame pointer */
+      /* For making room for guard value, scan all insns and fix the offset
+	 address of the variable that is based on frame pointer.
+	 Scan all declarations of variables and fix the offset address
+	 of the variable that is based on the frame pointer */
       sweep_string_variable (guard_area, UNITS_PER_GUARD);
 
 	
       /* the location of guard area moves to the beginning of stack frame */
       if ((offset = AUTO_OFFSET(XEXP (guard_area, 0))))
-	XEXP (XEXP (guard_area, 0), 1) = gen_rtx_CONST_INT (VOIDmode, sweep_frame_offset);
+	XEXP (XEXP (guard_area, 0), 1)
+	  = gen_rtx_CONST_INT (VOIDmode, sweep_frame_offset);
 
 
       /* Insert prologue rtl instructions */
@@ -192,9 +210,11 @@ prepare_stack_protection (inlinable)
 	  copy_args_for_protection ();
 
 #ifndef FRAME_GROWS_DOWNWARD
-	  /* If frame grows upward, character string copied from an arg stays top of
-	     the guard variable. So sweep the guard variable again */
-	  sweep_frame_offset = CEIL_ROUND (frame_offset, BIGGEST_ALIGNMENT / BITS_PER_UNIT);
+	  /* If frame grows upward, character string copied from an arg
+	     stays top of the guard variable.
+	     So sweep the guard variable again */
+	  sweep_frame_offset = CEIL_ROUND (frame_offset,
+					   BIGGEST_ALIGNMENT / BITS_PER_UNIT);
 	  sweep_string_variable (guard_area, UNITS_PER_GUARD);
 #endif
 	}
@@ -309,34 +329,36 @@ search_string_from_local_vars (block)
 	      && TREE_CODE (types) == VAR_DECL
 	      && ! DECL_ARTIFICIAL (types)
 	      && DECL_RTL_SET_P (types)
-	      && GET_CODE (DECL_RTL (types)) == MEM)
-	    {
-	      if (search_string_def (TREE_TYPE (types)))
-		{
-		  rtx home = DECL_RTL (types);
+	      && GET_CODE (DECL_RTL (types)) == MEM
 
-		  if (GET_CODE (home) == MEM
-		      && (GET_CODE (XEXP (home, 0)) == MEM
-			  || (GET_CODE (XEXP (home, 0)) == REG
-			      && XEXP (home, 0) != virtual_stack_vars_rtx
-			      && REGNO (XEXP (home, 0)) != HARD_FRAME_POINTER_REGNUM
-			      && REGNO (XEXP (home, 0)) != STACK_POINTER_REGNUM
+	      && search_string_def (TREE_TYPE (types)))
+	    {
+	      rtx home = DECL_RTL (types);
+
+	      if (GET_CODE (home) == MEM
+		  && (GET_CODE (XEXP (home, 0)) == MEM
+		      ||
+		      (GET_CODE (XEXP (home, 0)) == REG
+		       && XEXP (home, 0) != virtual_stack_vars_rtx
+		       && REGNO (XEXP (home, 0)) != HARD_FRAME_POINTER_REGNUM
+		       && REGNO (XEXP (home, 0)) != STACK_POINTER_REGNUM
 #if ARG_POINTER_REGNUM != HARD_FRAME_POINTER_REGNUM
-			      && REGNO (XEXP (home, 0)) != ARG_POINTER_REGNUM
+		       && REGNO (XEXP (home, 0)) != ARG_POINTER_REGNUM
 #endif
-			      )))
-		    /* If the value is indirect by memory or by a register
-		       that isn't the frame pointer
-		       then it means the object is variable-sized and address through
-		       that register or stack slot.  The protection has no way to hide pointer variables
-		       behind the array, so all we can do is staying the order of variables and arguments. */
-		    {
-		      current_function_has_variable_string = TRUE;
-		    }
-	    
-		  /* found character array */
-		  found = TRUE;
+		       )))
+		/* If the value is indirect by memory or by a register
+		   that isn't the frame pointer then it means the object is
+		   variable-sized and address through
+		   that register or stack slot.
+		   The protection has no way to hide pointer variables
+		   behind the array, so all we can do is staying
+		   the order of variables and arguments. */
+		{
+		  current_function_has_variable_string = TRUE;
 		}
+	    
+	      /* found character array */
+	      found = TRUE;
 	    }
 
 	  types = TREE_CHAIN(types);
@@ -382,17 +404,24 @@ search_string_def (type)
 	{
 	  /* Check if the string is a variable string */
 	  if (TYPE_DOMAIN (type) == 0
-	      || (TYPE_MAX_VALUE (TYPE_DOMAIN (type)) != 0
-		  && TREE_CODE (TYPE_MAX_VALUE (TYPE_DOMAIN (type))) == NOP_EXPR))
+	      ||
+	      (TYPE_MAX_VALUE (TYPE_DOMAIN (type)) != 0
+	       && TREE_CODE (TYPE_MAX_VALUE (TYPE_DOMAIN (type))) == NOP_EXPR))
 	    return TRUE;
 
 	  /* Check if the string size is greater than SUSPICIOUS_BUF_SIZE */
 	  if (TYPE_MAX_VALUE (TYPE_DOMAIN (type)) != 0
-	      && TREE_INT_CST_LOW(TYPE_MAX_VALUE(TYPE_DOMAIN(type)))+1 >= SUSPICIOUS_BUF_SIZE)
+	      &&
+	      TREE_INT_CST_LOW(TYPE_MAX_VALUE(TYPE_DOMAIN(type)))+1
+	      >= SUSPICIOUS_BUF_SIZE)
 	    return TRUE;
 
 	  current_function_defines_short_string = TRUE;
 	}
+      
+      /* to protect every functions, sweep any arrays to the frame top */
+      is_array = TRUE;
+
       return search_string_def(TREE_TYPE(type));
 	
     case UNION_TYPE:
@@ -508,15 +537,12 @@ search_pointer_def (type)
 	
     case POINTER_TYPE:
     case REFERENCE_TYPE:
-      /* I'm not sure whether OFFSET_TYPE needs this treatment,
-	 so I'll play safe and return 1.  */
     case OFFSET_TYPE:
       if (TYPE_READONLY (TREE_TYPE (type)))
 	{
-	  int funcp = search_func_pointer (TREE_TYPE (type), 1);
-	  /* Un-mark the type as having been visited already */
-	  search_func_pointer (TREE_TYPE (type), 0);
-	  return funcp;
+	  /* unless this pointer contains function pointer,
+	     it should be protected */
+	  return search_func_pointer (TREE_TYPE (type));
 	}
       return TRUE;
 	
@@ -529,9 +555,8 @@ search_pointer_def (type)
 
 
 static int
-search_func_pointer (type, mark)
+search_func_pointer (type)
      tree type;
-     int mark;
 {
   tree tem;
     
@@ -543,31 +568,37 @@ search_func_pointer (type, mark)
     case UNION_TYPE:
     case QUAL_UNION_TYPE:
     case RECORD_TYPE:
-	if (TREE_ASM_WRITTEN (type) != mark)
+	if (! TREE_VISITED (type))
 	  {
 	    /* mark the type as having been visited already */
-	    TREE_ASM_WRITTEN (type) = mark;
+	    TREE_VISITED (type) = 1;
 
 	    /* Output the name, type, position (in bits), size (in bits) of
 	       each field.  */
 	    for (tem = TYPE_FIELDS (type); tem; tem = TREE_CHAIN (tem))
 	      {
-		/* Omit here local type decls until we know how to support them. */
 		if (TREE_CODE (tem) == FIELD_DECL
-		    && search_func_pointer (TREE_TYPE(tem), mark)) return TRUE;
+		    && search_func_pointer (TREE_TYPE(tem))) {
+		  TREE_VISITED (type) = 0;
+		  return TRUE;
+		}
 	      }
+	    
+	    TREE_VISITED (type) = 0;
 	  }
 	break;
 
     case ARRAY_TYPE:
-      return search_func_pointer (TREE_TYPE(type), mark);
+      return search_func_pointer (TREE_TYPE(type));
 	
     case POINTER_TYPE:
     case REFERENCE_TYPE:
       /* I'm not sure whether OFFSET_TYPE needs this treatment,
 	 so I'll play safe and return 1.  */
     case OFFSET_TYPE:
-      return TREE_CODE (TREE_TYPE (type)) == FUNCTION_TYPE;
+      if (TREE_CODE (TREE_TYPE (type)) == FUNCTION_TYPE)
+	return TRUE;
+      return search_func_pointer (TREE_TYPE(type));
 	
     default:
       break;
@@ -729,7 +760,8 @@ rtl_prologue (insn)
   rtx _val;
 
   for (; insn; insn = NEXT_INSN (insn))
-    if (GET_CODE (insn) == NOTE && NOTE_LINE_NUMBER (insn) == NOTE_INSN_FUNCTION_BEG)
+    if (GET_CODE (insn) == NOTE
+	&& NOTE_LINE_NUMBER (insn) == NOTE_INSN_FUNCTION_BEG)
       break;
   
 #if !defined (HAS_INIT_SECTION)
@@ -741,13 +773,15 @@ rtl_prologue (insn)
     {
       rtx fbinsn = insn;
       for (; insn; insn = NEXT_INSN (insn))
-	if (GET_CODE (insn) == NOTE && NOTE_LINE_NUMBER (insn) == NOTE_INSN_BLOCK_BEG)
+	if (GET_CODE (insn) == NOTE
+	    && NOTE_LINE_NUMBER (insn) == NOTE_INSN_BLOCK_BEG)
 	  break;
       if (insn == 0) insn = fbinsn;
     }
 #endif
 
-  prologue_insert_point = NEXT_INSN (insn);	/* mark the next insn of FUNCTION_BEG insn */
+  /* mark the next insn of FUNCTION_BEG insn */
+  prologue_insert_point = NEXT_INSN (insn);
 		
   start_sequence ();
 
@@ -796,14 +830,16 @@ rtl_epilogue (insn)
     }
 #endif
 
-  compare_from_rtx (guard_area, _guard, NE, 0, GUARD_m, NULL_RTX);	/* if (guard_area != _guard) */
+  /*                                          if (guard_area != _guard) */
+  compare_from_rtx (guard_area, _guard, NE, 0, GUARD_m, NULL_RTX);
 
-  if_false_label = gen_label_rtx ();				/* { */
+  if_false_label = gen_label_rtx ();		/* { */
   emit_jump_insn ( gen_beq(if_false_label));
 
   /* generate string for the current function name */
-  funcstr = build_string (strlen(current_function_name)+1, current_function_name);
-  TREE_TYPE (funcstr) = build_array_type (char_type_node, 0);/* = char_array_type_node;*/
+  funcstr = build_string (strlen(current_function_name)+1,
+			  current_function_name);
+  TREE_TYPE (funcstr) = build_array_type (char_type_node, 0);
   funcname = output_constant_def (funcstr, 1);
 
   emit_library_call (gen_rtx (SYMBOL_REF, Pmode, "__stack_smash_handler"),
@@ -812,7 +848,7 @@ rtl_epilogue (insn)
 
   /* generate RTL to return from the current function */
 		
-  emit_barrier ();						/* } */
+  emit_barrier ();				/* } */
   emit_label (if_false_label);
 
   /* generate RTL to return from the current function */
@@ -854,43 +890,45 @@ arrange_var_order (block)
 	      && ! DECL_ARTIFICIAL (types)
 	      && ! DECL_INLINE (types)	/* don't sweep inlined string */
 	      && DECL_RTL_SET_P (types)
-	      && GET_CODE (DECL_RTL (types)) == MEM)
+	      && GET_CODE (DECL_RTL (types)) == MEM
+
+	      && (is_array=0, search_string_def (TREE_TYPE (types))
+		  || (! current_function_defines_vulnerable_string
+		      && is_array)))
 	    {
-	      if (search_string_def (TREE_TYPE (types)))
-		{
-		  rtx home = DECL_RTL (types);
+	      rtx home = DECL_RTL (types);
 
-		  if (! (GET_CODE (home) == MEM
-			 && (GET_CODE (XEXP (home, 0)) == MEM
-			     || (GET_CODE (XEXP (home, 0)) == REG
-				 && XEXP (home, 0) != virtual_stack_vars_rtx
-				 && REGNO (XEXP (home, 0)) != HARD_FRAME_POINTER_REGNUM
-				 && REGNO (XEXP (home, 0)) != STACK_POINTER_REGNUM
+	      if (!(GET_CODE (home) == MEM
+		    && (GET_CODE (XEXP (home, 0)) == MEM
+			||
+			(GET_CODE (XEXP (home, 0)) == REG
+			 && XEXP (home, 0) != virtual_stack_vars_rtx
+			 && REGNO (XEXP (home, 0)) != HARD_FRAME_POINTER_REGNUM
+			 && REGNO (XEXP (home, 0)) != STACK_POINTER_REGNUM
 #if ARG_POINTER_REGNUM != HARD_FRAME_POINTER_REGNUM
-				 && REGNO (XEXP (home, 0)) != ARG_POINTER_REGNUM
+			 && REGNO (XEXP (home, 0)) != ARG_POINTER_REGNUM
 #endif
-				 ))))
+			 ))))
+		{
+		  /* found a string variable */
+		  HOST_WIDE_INT var_size =
+		    ((TREE_INT_CST_LOW (DECL_SIZE (types)) + BITS_PER_UNIT - 1)
+		     / BITS_PER_UNIT);
+
+		  if (GET_MODE (DECL_RTL (types)) == BLKmode)
 		    {
-		      /* found a string variable */
-		      HOST_WIDE_INT var_size =
-			((TREE_INT_CST_LOW (DECL_SIZE (types)) + BITS_PER_UNIT - 1)
-			 / BITS_PER_UNIT);
-
-		      if (GET_MODE (DECL_RTL (types)) == BLKmode)
-			{
-			  int alignment = BIGGEST_ALIGNMENT / BITS_PER_UNIT;
-			  var_size = CEIL_ROUND (var_size, alignment);
-			}
-
-		      /* skip the variable if it is top of the region
-			 specified by sweep_frame_offset */
-		      offset = AUTO_OFFSET (XEXP (DECL_RTL (types), 0));
-		      if (offset == sweep_frame_offset - var_size)
-			sweep_frame_offset -= var_size;
-		      
-		      else if (offset < sweep_frame_offset - var_size)
-			sweep_string_variable (DECL_RTL (types), var_size);
+		      int alignment = BIGGEST_ALIGNMENT / BITS_PER_UNIT;
+		      var_size = CEIL_ROUND (var_size, alignment);
 		    }
+
+		  /* skip the variable if it is top of the region
+		     specified by sweep_frame_offset */
+		  offset = AUTO_OFFSET (XEXP (DECL_RTL (types), 0));
+		  if (offset == sweep_frame_offset - var_size)
+		    sweep_frame_offset -= var_size;
+		      
+		  else if (offset < sweep_frame_offset - var_size)
+		    sweep_string_variable (DECL_RTL (types), var_size);
 		}
 	    }
 
@@ -928,8 +966,9 @@ copy_args_for_protection ()
 	      {
 	        fntype = TREE_TYPE (current_function_decl);
 
-	        if ((TYPE_ARG_TYPES (fntype) != 0 &&
-	             TREE_VALUE (tree_last (TYPE_ARG_TYPES (fntype))) != void_type_node)
+	        if ((TYPE_ARG_TYPES (fntype) != 0
+		     && TREE_VALUE (tree_last (TYPE_ARG_TYPES (fntype))) 
+		          != void_type_node)
 	             || current_function_varargs)
 	          continue;
 	      }
@@ -943,6 +982,8 @@ copy_args_for_protection ()
 		int arg_size
 		  = ((TREE_INT_CST_LOW (DECL_SIZE (parms)) + BITS_PER_UNIT - 1)
 		     / BITS_PER_UNIT);
+		tree passed_type = DECL_ARG_TYPE (parms);
+		tree nominal_type = TREE_TYPE (parms);
 		
 		start_sequence ();
 
@@ -950,7 +991,8 @@ copy_args_for_protection ()
 		  {
 		    rtx safe = 0;
 		    
-		    change_arg_use_of_insns (prologue_insert_point, DECL_RTL (parms), &safe, 0);
+		    change_arg_use_of_insns (prologue_insert_point,
+					     DECL_RTL (parms), &safe, 0);
 		    if (safe)
 		      {
 			/* generate codes for copying the content */
@@ -963,7 +1005,6 @@ copy_args_for_protection ()
 			DECL_INCOMING_RTL (parms) = safe;
 		      }
 		  }
-
 		else if (GET_CODE (DECL_RTL (parms)) == MEM
 			 && GET_CODE (XEXP (DECL_RTL (parms), 0)) == ADDRESSOF)
 		  {
@@ -972,25 +1013,56 @@ copy_args_for_protection ()
 
 		    /* generate codes for copying the content */
 		    movinsn = emit_move_insn (safe, DECL_INCOMING_RTL (parms));
-		    PATTERN (movinsn)->volatil = 1;	/* avoid register elimination in gcse.c (COPY-PROP)*/
+		    /* avoid register elimination in gcse.c (COPY-PROP)*/
+		    PATTERN (movinsn)->volatil = 1;
 
-		    /* change the addressof information to the newly allocated pseudo register */
+		    /* change the addressof information to the newly
+		       allocated pseudo register */
 		    emit_move_insn (DECL_RTL (parms), safe);
 
 		    /* save debugger info */
 		    DECL_INCOMING_RTL (parms) = safe;
 		  }
 			
+		/* See if the frontend wants to pass this by invisible
+		   reference.  */
+		else if (passed_type != nominal_type
+			 && POINTER_TYPE_P (passed_type)
+			 && TREE_TYPE (passed_type) == nominal_type)
+		  {
+		    rtx safe = 0, orig = XEXP (DECL_RTL (parms), 0);
+
+		    change_arg_use_of_insns (prologue_insert_point,
+					     orig, &safe, 0);
+		    if (safe)
+		      {
+			/* generate codes for copying the content */
+			rtx movinsn = emit_move_insn (safe, orig);
+		    
+			/* avoid register elimination in gcse.c (COPY-PROP)*/
+			PATTERN (movinsn)->volatil = 1;
+			
+			/* save debugger info */
+			DECL_INCOMING_RTL (parms) = safe;
+		      }
+		  }
+
 		else
 		  {
-		    /* declare temporary local variable DECL_NAME (parms) for it */
+		    /* declare temporary local variable DECL_NAME (parms) */
 		    temp_rtx
 		      = assign_stack_local (DECL_MODE (parms), arg_size,
-					    DECL_MODE (parms) == BLKmode ? -1 : 0);
+					    DECL_MODE (parms) == BLKmode ?
+					    -1 : 0);
 		    
-		    MEM_IN_STRUCT_P (temp_rtx) = AGGREGATE_TYPE_P (TREE_TYPE (parms));
+		    MEM_IN_STRUCT_P (temp_rtx)
+		      = AGGREGATE_TYPE_P (TREE_TYPE (parms));
 		    set_mem_alias_set (temp_rtx, get_alias_set (parms));
 
+		    /* move_arg_location may change the contents of
+		       DECL_RTL (parms). to avoid this, copies the contents */
+		    /* SET_DECL_RTL (parms, copy_rtx (DECL_RTL (parms))); */
+
 		    /* generate codes for copying the content */
 		    store_expr (parms, temp_rtx, 0);
 
@@ -1026,8 +1098,8 @@ copy_args_for_protection ()
 
 
 /*
-  sweep a string variable to the local variable addressed by sweep_frame_offset, that is
-  a last position of string variables.
+  sweep a string variable to the local variable addressed
+  by sweep_frame_offset, that is a last position of string variables.
 */
 static void
 sweep_string_variable (sweep_var, var_size)
@@ -1053,17 +1125,21 @@ sweep_string_variable (sweep_var, var_size)
 
   /* scan all declarations of variables and fix the offset address of
      the variable based on the frame pointer */
-  sweep_string_in_decls (DECL_INITIAL (current_function_decl), sweep_offset, var_size);
+  sweep_string_in_decls (DECL_INITIAL (current_function_decl),
+			 sweep_offset, var_size);
 
-  /* scan all argument variable and fix the offset address based on the frame pointer */
-  sweep_string_in_args (DECL_ARGUMENTS (current_function_decl), sweep_offset, var_size);
+  /* scan all argument variable and fix the offset address based on
+     the frame pointer */
+  sweep_string_in_args (DECL_ARGUMENTS (current_function_decl),
+			sweep_offset, var_size);
 
-  /* For making room for sweep variable, scan all insns and fix the offset address
-     of the variable that is based on frame pointer*/
+  /* For making room for sweep variable, scan all insns and
+     fix the offset address of the variable that is based on frame pointer */
   sweep_string_use_of_insns (function_first_insn, sweep_offset, var_size);
 
 
-  /* Clear all the USED bits in operands of all insns and declarations of local vars */
+  /* Clear all the USED bits in operands of all insns and declarations of
+     local vars */
   reset_used_flags_for_decls (DECL_INITIAL (current_function_decl));
   reset_used_flags_for_insns (function_first_insn);
 
@@ -1080,12 +1156,13 @@ move_arg_location (insn, orig, new, var_size)
      rtx  insn, orig, new;
      HOST_WIDE_INT var_size;
 {
-  /* For making room for sweep variable, scan all insns and fix the offset address
-     of the variable that is based on frame pointer*/
+  /* For making room for sweep variable, scan all insns and
+     fix the offset address of the variable that is based on frame pointer */
   change_arg_use_of_insns (insn, orig, &new, var_size);
 
 
-  /* Clear all the USED bits in operands of all insns and declarations of local vars */
+  /* Clear all the USED bits in operands of all insns and declarations
+     of local vars */
   reset_used_flags_for_insns (insn);
 }
 
@@ -1128,13 +1205,17 @@ sweep_string_in_decls (block, sweep_offset, sweep_size)
 		  {
 		    offset = sweep_frame_offset - sweep_size - sweep_offset;
 
-		    XEXP (home, 0) = plus_constant (virtual_stack_vars_rtx, offset);
+		    XEXP (home, 0) = plus_constant (virtual_stack_vars_rtx,
+						    offset);
 		    XEXP (home, 0)->used = 1;
 		  }
 		else if (sweep_offset <= offset
 			 && offset < sweep_frame_offset)
-		  {	/* the rest of variables under sweep_frame_offset, so shift the location */
-		    XEXP (home, 0) = plus_constant (virtual_stack_vars_rtx, -sweep_size);
+		  {
+		    /* the rest of variables under sweep_frame_offset,
+		       shift the location */
+		    XEXP (home, 0) = plus_constant (virtual_stack_vars_rtx,
+						    -sweep_size);
 		    XEXP (home, 0)->used = 1;
 		  }
 	      }
@@ -1160,8 +1241,10 @@ sweep_string_in_decls (block, sweep_offset, sweep_size)
 			&& offset < sweep_offset + sweep_size)
 		      {
 
-			offset += sweep_frame_offset - sweep_size - sweep_offset;
-			XEXP (XEXP (home, 0), 1) = gen_rtx_CONST_INT (VOIDmode, offset);
+			offset
+			  += sweep_frame_offset - sweep_size - sweep_offset;
+			XEXP (XEXP (home, 0), 1) = gen_rtx_CONST_INT (VOIDmode,
+								      offset);
 
 			/* mark */
 			XEXP (home, 0)->used = 1;
@@ -1185,7 +1268,8 @@ sweep_string_in_decls (block, sweep_offset, sweep_size)
 	  types = TREE_CHAIN(types);
 	}
 
-      sweep_string_in_decls (BLOCK_SUBBLOCKS (block), sweep_offset, sweep_size);
+      sweep_string_in_decls (BLOCK_SUBBLOCKS (block),
+			     sweep_offset, sweep_size);
       block = BLOCK_CHAIN (block);
     }
 }
@@ -1217,15 +1301,19 @@ sweep_string_in_args (parms, sweep_offset, sweep_size)
 		    && offset < sweep_offset + sweep_size)
 		  {
 		    offset += sweep_frame_offset - sweep_size - sweep_offset;
-		    XEXP (XEXP (home, 0), 1) = gen_rtx_CONST_INT (VOIDmode, offset);
+		    XEXP (XEXP (home, 0), 1) = gen_rtx_CONST_INT (VOIDmode,
+								  offset);
 
 		    /* mark */
 		    XEXP (home, 0)->used = 1;
 		  }
 		else if (sweep_offset <= offset
 			 && offset < sweep_frame_offset)
-		  {	/* the rest of variables under sweep_frame_offset, so shift the location */
-		    XEXP (XEXP (home, 0), 1) = gen_rtx_CONST_INT (VOIDmode, offset - sweep_size);
+		  {
+		    /* the rest of variables under sweep_frame_offset,
+		       shift the location */
+		    XEXP (XEXP (home, 0), 1)
+		      = gen_rtx_CONST_INT (VOIDmode, offset - sweep_size);
 
 		    /* mark */
 		    XEXP (home, 0)->used = 1;
@@ -1248,8 +1336,10 @@ sweep_string_use_of_insns (insn, sweep_offset, sweep_size)
 	|| GET_CODE (insn) == CALL_INSN)
       {
 	has_virtual_reg = FALSE;
-	sweep_string_in_operand (insn, &PATTERN (insn), sweep_offset, sweep_size);
-	sweep_string_in_operand (insn, &REG_NOTES (insn), sweep_offset, sweep_size);
+	sweep_string_in_operand (insn, &PATTERN (insn),
+				 sweep_offset, sweep_size);
+	sweep_string_in_operand (insn, &REG_NOTES (insn),
+				 sweep_offset, sweep_size);
       }
 }
 
@@ -1332,7 +1422,9 @@ sweep_string_in_operand (insn, loc, sweep_offset, sweep_size)
 	    }
 	  else if (sweep_offset <= offset + k
 		   && offset + k < sweep_frame_offset)
-	    {	/* the rest of variables under sweep_frame_offset, so shift the location */
+	    {
+	      /* the rest of variables under sweep_frame_offset,
+		 shift the location */
 	      XEXP (x, 1) = gen_rtx_CONST_INT (VOIDmode, offset - sweep_size);
 	      x->used = 1;
 	    }
@@ -1410,16 +1502,14 @@ sweep_string_in_operand (insn, loc, sweep_offset, sweep_size)
 
 
 /*
-  change a argument variable to the local variable addressed by the "new" variable.
+  change a argument variable to the local variable addressed
+  by the "new" variable.
 */
-static int flag_caui_exit;
-
 static void
 change_arg_use_of_insns (insn, orig, new, size)
      rtx insn, orig, *new;
      HOST_WIDE_INT size;
 {
-  flag_caui_exit = FALSE;
   change_arg_use_of_insns_2 (insn, orig, new, size);
 }
 
@@ -1428,11 +1518,18 @@ change_arg_use_of_insns_2 (insn, orig, new, size)
      rtx insn, orig, *new;
      HOST_WIDE_INT size;
 {
-  for (; insn && !flag_caui_exit; insn = NEXT_INSN (insn))
+  for (; insn; insn = NEXT_INSN (insn))
     if (GET_CODE (insn) == INSN || GET_CODE (insn) == JUMP_INSN
 	|| GET_CODE (insn) == CALL_INSN)
       {
+	rtx seq;
+	
+	start_sequence ();
 	change_arg_use_in_operand (PATTERN (insn), orig, new, size);
+
+	seq = get_insns ();
+	end_sequence ();
+	emit_insn_before (seq, insn);
       }
 }
 
@@ -1494,8 +1591,9 @@ change_arg_use_in_operand (x, orig, new, size)
       
     case PLUS:
       /* Handle special case of frame register plus constant.  */
-      if (GET_CODE (orig) == MEM /* skip if orig is register variable in the optimization */
-	  && XEXP (x, 0) == virtual_incoming_args_rtx && CONSTANT_P (XEXP (x, 1))
+      if (GET_CODE (orig) == MEM
+	  && XEXP (x, 0) == virtual_incoming_args_rtx
+	  && CONSTANT_P (XEXP (x, 1))
 	  && ! x->used)
 	{
 	  offset = AUTO_OFFSET(x);
@@ -1537,35 +1635,25 @@ change_arg_use_in_operand (x, orig, new, size)
 	    offset = AUTO_OFFSET(XEXP (*new, 0))
 	      + (offset - AUTO_OFFSET(XEXP (orig, 0)));
 
-	    XEXP (x, 1) = plus_constant (virtual_stack_vars_rtx, offset);
+	    XEXP (x, 1) = force_operand (plus_constant (virtual_stack_vars_rtx,
+							offset), NULL_RTX);
 	    XEXP (x, 1)->used = 1;
 
-	  return;
+	    return;
 	  }
 	}
       break;
 
     case CALL_PLACEHOLDER:
-      change_arg_use_of_insns_2 (XEXP (x, 0), orig, new, size); if (flag_caui_exit) return;
-      change_arg_use_of_insns_2 (XEXP (x, 1), orig, new, size); if (flag_caui_exit) return;
-      change_arg_use_of_insns_2 (XEXP (x, 2), orig, new, size); if (flag_caui_exit) return;
+      change_arg_use_of_insns_2 (XEXP (x, 0), orig, new, size);
+      change_arg_use_of_insns_2 (XEXP (x, 1), orig, new, size);
+      change_arg_use_of_insns_2 (XEXP (x, 2), orig, new, size);
       break;
 
     default:
       break;
     }
 
-  if (*new == 0
-      && code == SET
-      && SET_SRC (x) == orig
-      && GET_CODE (SET_DEST (x)) == REG)
-    {
-      /* exit to the change_arg_use_of_insns */
-      flag_caui_exit = TRUE;
-      x->volatil = 1;	/* avoid register elimination in gcse.c (COPY-PROP)*/
-      return;
-    }
-  
   /* Scan all subexpressions.  */
   fmt = GET_RTX_FORMAT (code);
   for (i = 0; i < GET_RTX_LENGTH (code); i++, fmt++)
@@ -1661,7 +1749,7 @@ validate_operand_of_varrefs (insn, loc)
 	  && CONSTANT_P (XEXP (x, 1)))
 	{
 	  start_sequence ();
-	  /* temp = force_operand (x, NULL_RTX); */
+
 	  { /* excerpt from expand_binop in optabs.c */
 	    optab binoptab = add_optab;
 	    enum machine_mode mode = GET_MODE (x);
@@ -1671,8 +1759,8 @@ validate_operand_of_varrefs (insn, loc)
 	    rtx xop0 = XEXP (x, 0), xop1 = XEXP (x, 1);
 	    temp = gen_reg_rtx (mode);
 
-	    /* Now, if insn's predicates don't allow offset operands, put them into
-	       pseudo regs.  */
+	    /* Now, if insn's predicates don't allow offset operands,
+	       put them into pseudo regs.  */
 
 	    if (! (*insn_data[icode].operand[2].predicate) (xop1, mode1)
 		&& mode1 != VOIDmode)
@@ -1724,16 +1812,26 @@ validate_operand_of_varrefs (insn, loc)
   arbitrary memory locations.
 */
 #if !defined(FRAME_GROWS_DOWNWARD) && defined(STACK_GROWS_DOWNWARD)
-static void push_frame PARAMS ((HOST_WIDE_INT var_size, HOST_WIDE_INT boundary));
-static void push_frame_in_decls PARAMS ((tree block, HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
-static void push_frame_in_args PARAMS ((tree parms, HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
-static void push_frame_of_insns PARAMS ((rtx insn, HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
-static void push_frame_in_operand PARAMS ((rtx insn, rtx orig, HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
-static void push_frame_of_reg_equiv_memory_loc PARAMS ((HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
-static void push_frame_of_reg_equiv_constant PARAMS ((HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
+static void push_frame
+	PARAMS ((HOST_WIDE_INT var_size, HOST_WIDE_INT boundary));
+static void push_frame_in_decls
+	PARAMS ((tree block, HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
+static void push_frame_in_args
+	PARAMS ((tree parms, HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
+static void push_frame_of_insns
+	PARAMS ((rtx insn, HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
+static void push_frame_in_operand
+	PARAMS ((rtx insn, rtx orig,
+		 HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
+static void push_frame_of_reg_equiv_memory_loc
+	PARAMS ((HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
+static void push_frame_of_reg_equiv_constant
+	PARAMS ((HOST_WIDE_INT push_size, HOST_WIDE_INT boundary));
 static void reset_used_flags_for_push_frame PARAMS ((void));
-static int check_out_of_frame_access PARAMS ((rtx insn, HOST_WIDE_INT boundary));
-static int check_out_of_frame_access_in_operand PARAMS ((rtx, HOST_WIDE_INT boundary));
+static int check_out_of_frame_access
+	PARAMS ((rtx insn, HOST_WIDE_INT boundary));
+static int check_out_of_frame_access_in_operand
+	PARAMS ((rtx, HOST_WIDE_INT boundary));
 #endif
 
 rtx
@@ -1762,7 +1860,8 @@ assign_stack_local_for_pseudo_reg (mode, size, align)
   first_call_from_global_alloc = !saved_cse_not_expected && cse_not_expected;
   saved_cse_not_expected = cse_not_expected;
 
-  starting_frame = (STARTING_FRAME_OFFSET)?STARTING_FRAME_OFFSET:BIGGEST_ALIGNMENT / BITS_PER_UNIT;
+  starting_frame = (STARTING_FRAME_OFFSET)?
+    STARTING_FRAME_OFFSET:BIGGEST_ALIGNMENT / BITS_PER_UNIT;
   units_per_push = MAX(BIGGEST_ALIGNMENT / BITS_PER_UNIT,
 		       GET_MODE_SIZE (mode));
     
@@ -1805,7 +1904,8 @@ assign_stack_local_for_pseudo_reg (mode, size, align)
   
   if (push_frame_offset > push_allocated_offset)
     {
-      push_frame (units_per_push, push_allocated_offset + STARTING_FRAME_OFFSET);
+      push_frame (units_per_push,
+		  push_allocated_offset + STARTING_FRAME_OFFSET);
 
       assign_stack_local (BLKmode, units_per_push, -1);
       push_allocated_offset += units_per_push;
@@ -1836,13 +1936,18 @@ push_frame (var_size, boundary)
 {
   reset_used_flags_for_push_frame();
 
-  /* scan all declarations of variables and fix the offset address of the variable based on the frame pointer */
-  push_frame_in_decls (DECL_INITIAL (current_function_decl), var_size, boundary);
+  /* scan all declarations of variables and fix the offset address of
+     the variable based on the frame pointer */
+  push_frame_in_decls (DECL_INITIAL (current_function_decl),
+		       var_size, boundary);
 
-  /* scan all argument variable and fix the offset address based on the frame pointer */
-  push_frame_in_args (DECL_ARGUMENTS (current_function_decl), var_size, boundary);
+  /* scan all argument variable and fix the offset address based on
+     the frame pointer */
+  push_frame_in_args (DECL_ARGUMENTS (current_function_decl),
+		      var_size, boundary);
 
-  /* scan all operands of all insns and fix the offset address based on the frame pointer */
+  /* scan all operands of all insns and fix the offset address
+     based on the frame pointer */
   push_frame_of_insns (get_insns (), var_size, boundary);
 
   /* scan all reg_equiv_memory_loc and reg_equiv_constant*/
@@ -1859,7 +1964,8 @@ reset_used_flags_for_push_frame()
   extern rtx *reg_equiv_memory_loc;
   extern rtx *reg_equiv_constant;
 
-  /* Clear all the USED bits in operands of all insns and declarations of local vars */
+  /* Clear all the USED bits in operands of all insns and declarations of
+     local vars */
   reset_used_flags_for_decls (DECL_INITIAL (current_function_decl));
   reset_used_flags_for_insns (get_insns ());
 
@@ -1959,7 +2065,8 @@ push_frame_in_decls (block, push_size, boundary)
 		      && offset >= boundary)
 		    {
 		      offset += push_size;
-		      XEXP (XEXP (home, 0), 1) = gen_rtx_CONST_INT (VOIDmode, offset);
+		      XEXP (XEXP (home, 0), 1)
+			= gen_rtx_CONST_INT (VOIDmode, offset);
 		      
 		      /* mark */
 		      XEXP (home, 0)->used = 1;
@@ -2030,70 +2137,67 @@ push_frame_of_insns (insn, push_size, boundary)
     if (GET_CODE (insn) == INSN || GET_CODE (insn) == JUMP_INSN
 	|| GET_CODE (insn) == CALL_INSN)
       {
+	rtx last;
+	
 	insn_pushed = FALSE;
+
+	/* push frame in INSN operation */
 	push_frame_in_operand (insn, PATTERN (insn), push_size, boundary);
 
-	if (insn_pushed)
-	  {
-	    rtx trial = insn;
-	    rtx before = PREV_INSN (trial);
-	    rtx after = NEXT_INSN (trial);
-	    int has_barrier = 0;
-	    rtx tem;
-	    rtx seq = split_insns (PATTERN (insn), insn);
-
-	    /* If we are splitting a JUMP_INSN, it might be followed by a 
-	       BARRIER. We may need to handle this specially.  */
-	    if (after && GET_CODE (after) == BARRIER)
-	      {
-		has_barrier = 1;
-		after = NEXT_INSN (after);
-	      }
-
-	    if (seq && GET_CODE (seq) == SEQUENCE)
-	      {
-		if (XVECLEN (seq, 0) == 2)
-		  {
-		    rtx pattern = PATTERN (XVECEXP (seq, 0, 1));
-
-		    if (GET_CODE (pattern) == SET
-			&& GET_CODE (XEXP (pattern, 0)) == REG
-			&& GET_CODE (XEXP (pattern, 1)) == PLUS
-			&& XEXP (pattern, 0) == XEXP (XEXP (pattern, 1), 0)
-			&& CONSTANT_P (XEXP (XEXP (pattern, 1), 1)))
-		      {
-			rtx offset = XEXP (XEXP (pattern, 1), 1);
-			fp_equiv[REGNO (XEXP (pattern, 0))] = INTVAL (offset);
-
-			/* replace the pattern of the insn */
-			add_insn_after (XVECEXP (seq, 0, 0), before);
-			delete_insn (trial);
-			goto next;
-		      }
-		  }
-
-		/* excerpt from emit-rtl.c: L3320 */
-		tem = emit_insn_after (seq, trial);
-
-		delete_related_insns (trial);
-		if (has_barrier)
-		  emit_barrier_after (tem);
-
-		/* Recursively call try_split for each new insn created */
-		for (tem = NEXT_INSN (before); tem != after;
-		     tem = NEXT_INSN (tem))
-		  if (! INSN_DELETED_P (tem) && INSN_P (tem))
-		    tem = try_split (PATTERN (tem), tem, 1);
-	      }
-	  }
-
-      next:
 	/* push frame in NOTE */
 	push_frame_in_operand (insn, REG_NOTES (insn), push_size, boundary);
 
 	/* push frame in CALL EXPR_LIST */
 	if (GET_CODE (insn) == CALL_INSN)
-	  push_frame_in_operand (insn, CALL_INSN_FUNCTION_USAGE (insn), push_size, boundary);
+	  push_frame_in_operand (insn, CALL_INSN_FUNCTION_USAGE (insn),
+				 push_size, boundary);
+
+	if (insn_pushed
+	    && (last = try_split (PATTERN (insn), insn, 1)) != insn)
+	  {
+	    rtx first = NEXT_INSN (insn);
+	    rtx trial = NEXT_INSN (first);
+	    rtx pattern = PATTERN (trial);
+	    rtx set;
+
+	    /* update insn_chain with splitted insns
+	       when it is called from global_alloc. */
+	    if (cse_not_expected)
+	      {
+		struct insn_chain *chain = reload_insn_chain;
+		for (; chain; chain = chain->next)
+		  if (chain->insn == insn)
+		    {
+		      chain->insn = first;
+		      break;
+		    }
+	      }
+
+	    /* update REG_EQUIV info to the first splitted insn */
+	    if ((set = single_set (insn))
+		&& find_reg_note (insn, REG_EQUIV, SET_SRC (set))
+		&& GET_CODE (PATTERN (first)) == SET)
+	      {
+		REG_NOTES (first)
+		  = gen_rtx_EXPR_LIST (REG_EQUIV,
+				       SET_SRC (PATTERN (first)),
+				       REG_NOTES (first));
+	      }
+
+	    if (GET_CODE (pattern) == SET
+		&& GET_CODE (XEXP (pattern, 0)) == REG
+		&& GET_CODE (XEXP (pattern, 1)) == PLUS
+		&& XEXP (pattern, 0) == XEXP (XEXP (pattern, 1), 0)
+		&& CONSTANT_P (XEXP (XEXP (pattern, 1), 1)))
+	      {
+		rtx offset = XEXP (XEXP (pattern, 1), 1);
+		fp_equiv[REGNO (XEXP (pattern, 0))] = INTVAL (offset);
+
+		delete_insn (trial);
+	      }
+
+	    insn = last;
+	  }
       }
 
   /* Clean up.  */
